#!/usr/bin/env python

# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

import avroknife

setup(
    name=avroknife.__title__,
    version=avroknife.__version__,
    author=avroknife.__author__,
    author_email=avroknife.__email__,
    packages=find_packages(),
    include_package_data = True,
    url='https://github.com/CeON/avroknife',
    license=avroknife.__license__,
    description=avroknife.__description__,
    ## This file is generated by running appropriate target from the Makefile
    long_description=open('tmp/README.rst').read(),
    install_requires=[
        'avro>=1.7.5',
        'nose'
    ],
    ## Add ability to run tests in the code through the setup.py script
    test_suite='nose.collector',
    keywords=['Avro', 'browsing', 'manipulation'],
    classifiers=[
        'Development Status :: 5 - Production/Stable    ',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Operating System :: POSIX :: Linux',
    ],
    scripts=['scripts/avroknife']
)
